import {inject} from 'aurelia-framework';
import {HttpClient} from 'aurelia-http-client';
import {Router} from 'aurelia-router';
import environment from 'environment';

@inject(Router)
export class KaApi {
  routeLogin = 'login';
  routeHome = 'home';
  storage = environment.storage && window[environment.storage] ? window[environment.storage] : localStorage;
  _accessToken = '';
  _refreshToken = '';
  _refreshPromise = null;
  _user = null;
  _baseUrl = this.storage.getItem('api_base_url') || (environment.api && environment.api.kamaji ? environment.api.kamaji : environment.api);
  _isKamaji1 = environment.api && typeof environment.api === 'string' && !!environment.api.match(/api\.kamaji\.it\/v1\//g); // backwards compatibility

  constructor(router) {
    this._accessToken = this.storage.getItem('access_token') || '';
    this._refreshToken = this.storage.getItem('refresh_token') || '';
    this._user = this.storage.getItem('user') ? JSON.parse(this.storage.getItem('user')) : null;
    this._refreshUrl = '';
    this.client = new HttpClient();
    this.client._baseUrl = this._baseUrl; // Keep reference of baseUrl into HttpClient object
    this.router = router;

    this.client.configure(x => {
      x.withBaseUrl(this.baseUrl);
      x.withInterceptor({
        request: msg => {
          if (this.accessToken.length && (!this._refreshUrl.length || (msg.url !== this._refreshUrl))) {
            msg.headers.add('Authorization', 'Bearer ' + this.accessToken);
          }
          return msg;
        },
        response: msg => {
          if (String(msg.statusCode) !== '204' && msg.responseType === 'json' && typeof msg.response === 'string') {
            msg.response = JSON.parse(msg.response);
          }
          return msg;
        },
        responseError: msg => {
          /*
          if (msg.statusCode === 0) {
            return new Promise(() => {
              setTimeout(() => {
                return this.client.send(msg.requestMessage);
              }, 2000);
            });
          }
          */
          if (msg.statusCode === 401 && this.refreshToken.length && msg.requestMessage.url !== this._refreshUrl) {
            return this.refresh(this.refreshToken).then(success => {
              return this.client.send(msg.requestMessage);
            }, error => {
              return error;
            });
          }
          return Promise.reject(msg);
        }
      });
    });
  }

  get baseUrl() {
    return this._baseUrl;
  }
  set baseUrl(value) {
    this._baseUrl = value;
    this.storage.setItem('api_base_url', value);
    if (this.client) {
      this.client._baseUrl = this._baseUrl;
      this.client.configure(conf => { conf.withBaseUrl(value); });
    }
  }

  get accessToken() {
    return this._accessToken;
  }
  set accessToken(value) {
    this._accessToken = value || '';
    if (value) this.storage.setItem('access_token', value);
    else this.storage.removeItem('access_token');
  }

  get refreshToken() {
    return this._refreshToken;
  }
  set refreshToken(value) {
    this._refreshToken = value || '';
    if (value) this.storage.setItem('refresh_token', value);
    else this.storage.removeItem('refresh_token');
  }

  get user() {
    return this._user;
  }
  set user(value) {
    this._user = value || '';
    if (value) this.storage.setItem('user', JSON.stringify(value));
    else this.storage.removeItem('user');
  }

  get isAuthenticated() {
    return !!this.accessToken.length && !!this.user;
  }

  login(username, password) {
    return this.client.post('auth', { username, password, grant_type: 'password' }).then(msg => {
      this.accessToken = msg.response.access_token;
      this.refreshToken = msg.response.refresh_token;
      if (this._isKamaji1) {
        return this.client.get('auth').then(x => {
          let response = x.response;
          return this.client.get(`${response.endpoint}/${response.userid}`, { depth: 0 }).then(x => {
            this.user = x.response;
            konsole.debug(this.user);
            this.router.navigateToRoute(this.routeHome);
            return Promise.resolve(x);
          }, x => {
            this.user = null;
            return Promise.reject(x);
          });
        }, x => {
          this.user = null;
          return Promise.reject(x);
        });
      }
      return this.client.get('me?depth=0').then(x => {
        this.user = x.response;
        konsole.debug('me', this.user);
        this.router.navigateToRoute(this.routeHome);
        return Promise.resolve(x);
      }, x => {
        this.user = null;
        return Promise.reject(x);
      });
    }, x => {
      this.accessToken = '';
      this.refreshToken = '';
      this.user = null;
      return Promise.reject(x);
    });
  }

  logout() {
    this.accessToken = '';
    this.refreshToken = '';
    this.user = null;
    localStorage.clear();
    sessionStorage.clear();
    this.router.navigateToRoute(this.routeLogin);
  }

  refresh(token) {
    if (this._refreshPromise) return this._refreshPromise;
    this._refreshUrl = 'auth/refresh/' + token;
    this._refreshPromise = this.client.post(this._refreshUrl).then(msg => {
      this.accessToken = msg.response.access_token;
      this.refreshToken = msg.response.refresh_token;
      return Promise.resolve(msg.response);
    }, x => {
      this.logout();
      return Promise.reject(x);
    });
    this._refreshPromise.finally(() => {
      this._refreshUrl = '';
      this._refreshPromise = null;
    });
    return this._refreshPromise;
  }

  get(endpoint, params) {
    let queryString = (typeof params === 'object') ? '?' + this.buildQueryString(params) : '';
    return this.client.get(this.baseUrl + endpoint + queryString);
  }

  post(endpoint, data) {
    return this.client.post(this.baseUrl + endpoint, data);
  }

  put(endpoint, data) {
    return this.client.put(this.baseUrl + endpoint, data);
  }

  patch(endpoint, data) {
    return this.client.patch(this.baseUrl + endpoint, data);
  }

  delete(endpoint) {
    return this.client.delete(this.baseUrl + endpoint);
  }

  buildQueryString(params) {
    let query = [];
    for (let [key, value] of Object.entries(params)) query.push(`${key}=` + encodeURIComponent(value));
    return query.join('&');
  }
}

