import {inject, customElement, bindable, bindingMode, BindingEngine, InlineViewStrategy, NewInstance} from 'aurelia-framework';
import {ValidationController, ValidationRules, validateTrigger} from 'aurelia-validation';
import {Router} from 'aurelia-router';
import {KaApi} from 'ka-components';


@customElement('ka-control')
@inject(Element, Router, KaApi, BindingEngine, NewInstance.of(ValidationController))
export class KaControl {
  // Basic input control properties
  @bindable() schema = null;
  @bindable({defaultBindingMode: bindingMode.twoWay}) value = null;
  // Other input control bindable properties
  @bindable() name = null;
  @bindable() readonly = null;
  @bindable() required = null;
  @bindable() label = null;
  @bindable() description = null;
  @bindable() placeholder = '';
  @bindable() error = null;
  // Input control specific params property
  @bindable() params = null;
  // Other control input properties
  errors = null;
  buttons = [];
  // Bindable callbacks
  @bindable() change = (value) => {};

  constructor(element, router, api, binding, validator) {
    this.element = element;
    this.router = router;
    this.api = api;
    this.binding = binding;
    this.validator = validator;
    this.validator.validateTrigger = validateTrigger.manual; // This is because we don't want model to be automatically validated on input blur

    ValidationRules
      .ensure('value')
      .required()
      .when(self => self.required)
      .withMessage('Campo obbligatorio')
      .equals(true)
      .when(self => self.schema.control === 'check' && self.required)
      .withMessage('Spunta obbligatoria')

      .satisfies((value, self) => value >= self.schema.min && value <= self.schema.max)
      .when(self => self.schema.control === 'number' && self.schema.min && self.schema.max)
      .withMessage('Il valore deve essere compreso tra ${min} e ${max}')
      .satisfies((value, self) => value >= self.schema.min)
      .when(self => self.schema.control === 'number' && self.schema.min && !self.schema.max)
      .withMessage('Il valore deve essere uguale o maggiore di ${min}')
      .satisfies((value, self) => value <= self.schema.max)
      .when(self => self.schema.control === 'number' && !self.schema.min && self.schema.max)
      .withMessage('Il valore deve essere minore o uguale a ${max}')

      .satisfies((value, self) => value.length >= self.schema.min && value.length <= self.schema.max)
      .when(self => self.value && ['text', 'textarea', 'password'].includes(self.schema.control) && self.schema.min && self.schema.max)
      .withMessage('Il numero di caratteri deve essere compreso tra ${min} e ${max}')
      .satisfies((value, self) => value.length >= self.schema.min)
      .when(self => self.value && ['text', 'textarea', 'password'].includes(self.schema.control) && self.schema.min && !self.schema.max)
      .withMessage('Il valore deve essere di almeno ${min} caratteri')
      .satisfies((value, self) => value.length <= self.schema.max)
      .when(self => self.value && ['text', 'textarea', 'password'].includes(self.schema.control) && !self.schema.min && self.schema.max)
      .withMessage('Il valore non deve superare i ${max} caratteri')

      .ensure('displayValue')
      .required()
      .when(self => self.required)
      .withMessage('Campo obbligatorio')
      .satisfies((value, self) => value === null || value === '' || ((/^\d{2}\/\d{2}\/\d{4}$/g).test(value) && moment(value, 'DD-MM-YYYY').isValid()))
      .when(self => self.schema.control === 'date' || self.schema.control === 'age')
      .withMessage('Formato data non valido')
      .on(this);
  }

  bind(bindingContext) {
    // Reference to view-model containing this component
    this.parentContext = bindingContext;
    // Bind to ka-resource if any in parentContext
    if (this.name && this.parentContext.resource) this.bindResource();
    // Handle control if not attached to ka-resource
    if (!this.name && this.schema) this.schemaChanged(this.schema);
  }

  attached() {
    if (this.element.hasAttribute('buttons')) {
      this.buttons = this.element.getAttribute('buttons').split(',');
    }
  }

  detached() {
    this.unbindResource();
  }

  schemaChanged(schema) {
    if (!schema) return;
    this.readonly = this.readonly !== null ? String(this.readonly) === 'true' : String(schema.readonly) === 'true';
    if (!this.readonly && this._bindResource && this.parentContext.resource.readonly !== null) this.readonly = String(this.parentContext.resource.readonly) === 'true';
    this.required = this.required !== null ? String(this.required) === 'true' : String(schema.required) === 'true';
    this.label = this.label || schema.label || null;
    this.placeholder = this.placeholder || schema.placeholder || '';
    this.description = this.description || schema.description || null;
    this.viewStrategy = new InlineViewStrategy(`<template><ka-control-${schema.control} schema.bind="schema" value.bind="value & validate" readonly.bind="readonly" placeholder.bind="placeholder" ${schema.control === 'check' ? 'description.bind="description"' : ''} params.bind="params"></ka-control-${schema.control}></template>`);
  }
  valueChanged(value) {
    if (this._bindResource && this.parentContext.resource.data) this.parentContext.resource.data[this.name] = value;
    this.validate();
    if (typeof this.change === 'function') {
      this.change(value);
    }
  }
  readonlyChanged() {
    if (!this.schema) return;
    if (String(this.schema.readonly) === 'true') this.readonly = true;
  }

  validate() {
    let validator = this.validator.validate();
    validator.then(result => {
      if (result.valid) {
        this.element.classList.remove('error');
      } else {
        this.element.classList.add('error');
      }
    });
    return validator;
  }
  setError(message) {
    if (message && typeof message === 'string') {
      this.element.classList.add('error');
      this.validator.errors = [message];
    } else if (message) {
      this.element.classList.add('error');
      this.validator.errors = message;
    } else {
      this.element.classList.remove('error');
      this.validator.errors = [];
    }
  }

  bindResource() {
    if (this.parentContext.resource) {
      this.parentContext.resource.controls[this.name] = this;
      if (this.parentContext.resource.schema) {
        this.resourceSchemaBinding = this.binding.expressionObserver(this, `parentContext.resource.schema.${this.name}`).subscribe((value) => this.schema = value );
      } else if (this.schema) {
        this.schemaChanged(this.schema);
      }
      if (this.parentContext.resource.data) {
        this.resourceValueBinding = this.binding.expressionObserver(this, `parentContext.resource.data.${this.name}`).subscribe((value) => { if (value !== this.value) this.value = value; });
      }
    }
    if (!this.readonly) this.resourceReadonlyBinding = this.binding.expressionObserver(this, 'parentContext.resource.readonly').subscribe((value) => { this.readonly = value; });
    this._bindResource = true;
  }
  unbindResource() {
    if (this.resourceSchemaBinding) this.resourceSchemaBinding.dispose();
    if (this.resourceValueBinding) this.resourceValueBinding.dispose();
    if (this.resourceReadonlyBinding) this.resourceReadonlyBinding.dispose();
    this._bindResource = false;
  }

  clear() {
    this.value = null;
  }
}
