import {inject, customElement, containerless, BindingEngine, bindable, bindingMode, NewInstance} from 'aurelia-framework';
import {Router} from 'aurelia-router';
import {ValidationController, ValidationRules, validateTrigger} from 'aurelia-validation';
import {HttpClient} from 'aurelia-http-client';
import {KaDialog} from '../ka-dialog/ka-dialog';
import ClassicEditor from 'ka-ckeditor';
import moment from 'moment';

@customElement('ka-input-controls')
@containerless()
@inject(Element, Router, NewInstance.of(ValidationController), BindingEngine, KaDialog)
export class KaInputControls {

  @bindable({defaultBindingMode: bindingMode.twoWay}) value = null;
  @bindable({defaultBindingMode: bindingMode.twoWay}) displayValue = null;
  @bindable({defaultBindingMode: bindingMode.twoWay}) objectValue = null;
  // Kamaji's field configuration
  @bindable() config = null;
  // Reference to record id
  @bindable() record = null;
  // Field configuration overrides
  @bindable() name = null;
  @bindable() label = null;
  @bindable() default = null;
  @bindable() required = null;
  @bindable() readonly = null;
  @bindable() control = null;
  @bindable() client = null;
  @bindable() description = null;
  // Other parameters
  @bindable() validation = true;
  @bindable() queryOperator = null; // operator for combo queries
  // Event callbacks
  @bindable() change = null;
  @bindable() ready = null;
  @bindable() dialogClosed = null;

  constructor(element, router, validator, bindingEngine, dialog) {
    this.element = element;
    this.router = router;
    this.bindingEngine = bindingEngine;
    this.dialog = dialog;
    this.validator = validator;
    this.validator.validateTrigger = validateTrigger.manual; // This is because we don't want model to be automatically validated on input blur
    this._client = new HttpClient(); // Internal http client, to use with combo fields with datasource = url, or with a combo field for which a client is not specified
    this._client.configure(x => {
      x.withInterceptor({
        response: msg => {
          if (msg.responseType === 'json' && typeof msg.response === 'string') {
            msg.response = JSON.parse(msg.response);
          }
          return msg;
        }
      });
    });
    // SET VALIDATION RULES
    ValidationRules
      .ensure('value')
        .required()
          .when(self => self.required)
          .withMessage('Campo obbligatorio')
        .equals(true)
          .when(self => self.control === 'check' && self.required)
          .withMessage('Spunta obbligatoria')

        .satisfies((value, self) => value >= self.min && value <= self.max)
          .when(self => self.control === 'number' && self.min && self.max)
          .withMessage('Il valore deve essere compreso tra ${min} e ${max}')
        .satisfies((value, self) => value >= self.min)
          .when(self => self.control === 'number' && self.min && !self.max)
          .withMessage('Il valore deve essere uguale o maggiore di ${min}')
        .satisfies((value, self) => value <= self.max)
          .when(self => self.control === 'number' && !self.min && self.max)
          .withMessage('Il valore deve essere minore o uguale a ${max}')

        .satisfies((value, self) => value.length >= self.min && value.length <= self.max)
          .when(self => self.value && ['text', 'textarea', 'password'].includes(self.control) && self.min && self.max)
          .withMessage('Il numero di caratteri deve essere compreso tra ${min} e ${max}')
        .satisfies((value, self) => value.length >= self.min)
          .when(self => self.value && ['text', 'textarea', 'password'].includes(self.control) && self.min && !self.max)
          .withMessage('Il valore deve essere di almeno ${min} caratteri')
        .satisfies((value, self) => value.length <= self.max)
          .when(self => self.value && ['text', 'textarea', 'password'].includes(self.control) && !self.min && self.max)
          .withMessage('Il valore non deve superare i ${max} caratteri')

      .ensure('displayValue')
        .required()
          .when(self => self.required)
          .withMessage('Campo obbligatorio')
        .satisfies((value, self) => value === null || value === '' || ((/^\d{2}\/\d{2}\/\d{4}$/g).test(value) && moment(value, 'DD-MM-YYYY').isValid()))
          .when(self => self.control === 'date' || self.control === 'age')
          .withMessage('Formato data non valido')
    .on(this);
  }

  bind(bindingContext, overrideContext) {
    this.parentContext = bindingContext;
  }
  attached() {
    this.init();
  }

  init() {
    if (!this.initialized) {
      if (!this.config || !this.element || !this.element.parentElement) return;
      this.element = this.element.parentElement;
    }
    //konsole.debug('ka-input initialized' + (this.initialized ? ' again' : ''), this.config.name || this.config.field);
    this.initialized = true;
    
    let config_overrides = {
      label: this.label !== null ? this.label : (this.config && this.config.label ? this.config.label : null),
      default: this.default !== null ? this.default : (this.config && this.config.default ? this.config.default : null),
      required: this.required !== null ? this.required === 'true' || this.required === true : (this.config && this.config.required ? this.config.required : null),
      readonly: this.readonly !== null ? this.readonly === 'true' || this.readonly === true : (this.config && this.config.readonly ? this.config.readonly : null),
      control: this.control !== null ? this.control : (this.config && this.config.control ? this.config.control : null)
    };
    Object.assign(this, this.config, config_overrides);
    this.config = null;

    this.id = 'ka-input-' + (this.field || 'noname') + '-' + Math.random().toString(36).slice(-6);

    // Parse datasource
   
    if (typeof this.datasource === 'string' && (this.datasource.indexOf('{') === 0 || this.datasource.indexOf('[') === 0)) {
      this.datasource = JSON.parse(this.datasource);
      if (this.datasource.table && this.datasource.table.indexOf('_') === 0) this.datasource.table = this.datasource.table.substring(1); // Kamaji1 backwards compatibility
    }
    if (this.datasource && !this.datatext)
      this.datatext = 'text';
    if (this.datasource && !this.datavalue)
      this.datavalue = 'value';

    // Parse value
    if (!this.value && this.default)
      this.value = this.default;
    if (this.value && typeof this.value === 'object' && this.datasource && this.datavalue && this.value[this.datavalue])
      this.value = String(this.value[this.datavalue]);
    if (this.value && typeof this.value === 'object' && this.datatype === 'json')
      this.value = JSON.stringify(this.value);

    // INIT INPUT
    let input;
    switch (this.control) {
    case 'text':
      this.type = 'text';
      input = this.element.querySelector('input');
      if (input && this.min) input.setAttribute('minlength', this.min);
      if (input && this.max) input.setAttribute('maxlength', this.max);
      break;
    case 'textarea':
      this.type = 'textarea';
      input = this.element.querySelector('textarea');
      if (input && this.min) input.setAttribute('minlength', this.min);
      if (input && this.max) input.setAttribute('maxlength', this.max);
      break;
    case 'password':
      this.type = 'password';
      input = this.element.querySelector('input');
      if (input && this.min) input.setAttribute('minlength', this.min);
      if (input && this.max) input.setAttribute('maxlength', this.max);
      break;
    case 'number':
      this.type = 'number';
      input = this.element.querySelector('input');
      if (input && this.min) input.setAttribute('min', this.min);
      if (input && this.max) input.setAttribute('max', this.max);
      break;
    case 'check':
      this.type = !this.datamultiple ? 'checkbox' : 'checklist';
      if (this.datamultiple && this.datapreload) this.checklistInit();
      if (this.datamultiple && !this.datapreload) this.comboListUpdate();
      break;
    case 'date':
    case 'age':
      this.type = 'text';
      this.placeholder = this.placeholder || 'gg/mm/aaaa';
      this.dateInit();
      break;
    case 'time':
      this.type = 'time';
      this.placeholder = this.placeholder || 'HH:mm';
      break;
    case 'combo':
      this.type = 'combo';
      this.comboInit();
      break;
    case 'editor':
      this.type = 'editor';
      this.editorInit();
      break;
    case 'file':
      this.type = 'file';
      break;
    default:
      this.type = 'text';
      break;
    }
    this.element.classList.add('type-' + this.type);

    // Parse static attributes
    if (this.element.hasAttribute('description')) this.description = this.element.getAttribute('description');
    else if (this.control === 'check' && !this.datamultiple && !this.description) { this.description = this.label; this.label = null; }
    if (this.element.hasAttribute('placeholder')) this.placeholder = this.element.getAttribute('placeholder');
    else this.placeholder = this.placeholder || '';

    // Other classes
    if (this.readonly) this.element.classList.add('readonly');
    if (this.required) this.element.classList.add('required');

    // INIT buttons

    this.buttons = {};
    if (!this.element.classList.contains('no-buttons')) {
      if (!this.element.classList.contains('no-clear') && this.control !== 'editor' && this.control !== 'check') this.buttons.clearEnabled = true;
      if (!this.element.classList.contains('no-add') && this.control === 'combo' && this.datasource && this.datasource.table && !this.datamultiple) this.buttons.addEnabled = !this.value;
      if (!this.element.classList.contains('no-open') && this.control === 'combo' && this.datasource && this.datasource.table && !this.datamultiple) this.buttons.openEnabled = !!this.value;
      if (this.control === 'password') this.buttons.showEnabled = true;
      if (this.control === 'file') this.buttons.downloadEnabled = true;
    }
    this.buttons.length = Object.keys(this.buttons).length;

    this.buttons.clear = () => {
      this.value = null;
      if (this.control === 'date' || this.control === 'age') this.dateDisplayValue();
      if (this.control === 'combo') this.comboDisplayValue();
      this.validate();
    };
    this.buttons.open = () => {
      let childRoute = this.router.routes.filter(x => { return x.name == (this.datasource.table || this.router.currentInstruction.config.name) })[0];
      if (!childRoute || !this.value || isNaN(parseInt(this.value, 10))) return;
      this.dialog.open({
        title: childRoute.title,
        viewModel: childRoute.moduleId,
        viewModelParams: { id: this.value }
      }, true).whenClosed(response => {
        if (this.control == 'combo') {
          this.comboItemsSource = null;
          this.comboDisplayValue();
        }
        if (typeof this.dialogClosed === 'function') this.dialogClosed(response);
      });
    };
    this.buttons.add = () => {
      let childRoute = this.router.routes.filter(x => { return x.name == (this.datasource.table || this.router.currentInstruction.config.name) })[0];
      if (!childRoute || this.value) return;
      this.dialog.open({
        title: childRoute.title,
        viewModel: childRoute.moduleId,
        viewModelParams: { id: 'new' }
      }, true).whenClosed(response => {
        if (response.output && response.output.id) {
          this.value = response.output.id;
        }
        if (typeof this.dialogClosed === 'function') this.dialogClosed(response);
      });
    };
    this.buttons.show = () => {
      this.type = 'text';
      this.buttons.showEnabled = false;
      this.buttons.hideEnabled = true;
    };
    this.buttons.hide = () => {
      this.type = 'password';
      this.buttons.showEnabled = true;
      this.buttons.hideEnabled = false;
    };
    this.buttons.download = () => {
      if (!this.value) return;
      let download = (this.element.hasAttribute('download') && this.element.getAttribute('download') === 'false') ? 0 : 1;
      window.open(`${this.client._baseUrl}storage/${this.record}/${this.field}?download=${download}&access_token=${localStorage.getItem('access_token') || sessionStorage.getItem('access_token')}`);
    };
    if (this.control === 'combo') {
    } else {
      this._ready = true;
      if (typeof this.ready === 'function') this.ready(this);
      this.format();
    }
  }
  configChanged(value) {
    if (value) this.init();
  }
  valueChanged(value) {
    if (this.control === 'date' || this.control === 'age') {
      this.dateDisplayValue();
    }
    if (this.control === 'combo') {
      this.comboDisplayValue();
    }
    if (this.control === 'check' && this.datamultiple) {
      if (this.datapreload) this.checklistParseSelected();
      else this.comboListUpdate();
    }
    if (this.control === 'editor' && this.editor && this.editor.data) {
      if (this.editor.data.get() !== value) this.editor.data.set(value || '');
    }
    if (this.control === 'file') {
      this.fileDisplayValue();
    }
    if (this.value && typeof this.value === 'object' && this.datatype === 'json') {
      this.value = JSON.stringify(this.value);
    }
    if (typeof this.change === 'function' && this._ready) {
      setTimeout(() => {
        this.change({ value: this.value });
      }, 100);
    }
  }
  requiredChanged(value) {
    if (!this.initialized) return;
    if (value) this.element.classList.add('required');
    else this.element.classList.remove('required');
  }
  readonlyChanged(value) {
    if (!this.initialized) return;
    if (value) this.element.classList.add('readonly');
    else this.element.classList.remove('readonly');
    if (this.control === 'editor' && this.editor) this.editor.isReadOnly = value;
  }
  focus() {
    if (this.readonly) return;
    this.element.classList.add('focus');
  }
  blur() {
    if (this.readonly) return;
    this.element.classList.remove('focus');
    this.validate();
  }
  keyup($event) {
    if (this.control === 'date' || this.control === 'age') {
      if (!this.displayValue || $event.keyCode === 8 || $event.keyCode === 9 || $event.keyCode === 16 || $event.keyCode === 46) return; // Break on backspace + delete
      let value = this.displayValue.split('-');
      value = (value[2] ? value[2] + '/' : '') + (value[1] ? value[1] + '/' : '') + (value[0] ? value[0] : '');
      if(!(/^([0123]|(?:0[1-9])|[12]\d|3[01])$|^(\d{2})\/$|^(\d{2})\/([01]|(?:0[1-9])|1[012])$|^(\d{2})\/(\d{2})\/$|^(\d{2})\/(\d{2})\/(\d{0,4})$/g).test(value))
        value = value.slice(0, -1);
      if ((/^\d{2}$|^\d{2}\/\d{2}$/g).test(value))
        value += '/';
      this.displayValue = value;
    }
  }
  format() {
    if (typeof this.value === 'undefined') return;
    switch (this.control) {
    case 'text':
      this.value = this.value;
      break;
    case 'password':
      this.value = !this.value || this.value.length === 0 ? null : this.value;
      break;
    case 'number':
      this.value = this.value;
      break;
    case 'date':
    case 'age':
      this.value = !this.displayValue || this.displayValue === '' || this.displayValue === '0000-00-00' ? null : (/^\d{2}\/\d{2}\/\d{4}$/g).test(this.displayValue) && moment(this.displayValue, 'DD/MM/YYYY').isValid() ? moment(this.displayValue, 'DD/MM/YYYY').format('YYYY-MM-DD') : this.displayValue;
      break;
    case 'combo':
      this.value = this.value;
      break;
    case 'file':
      this.value = this.value;
      break;
    default:
      this.value = this.value;
      break;
    }
  }
  validate() {
    this.format();
    if (!this.validation) {
      return new Promise((resolve) => {
        this.element.classList.remove('error');
        resolve({
          instruction: undefined,
          results: [],
          valid: true
        });
      });
    }
    return new Promise((resolve, reject) => {
      this.validator.validate().then(result => {
        if (result.valid) {
          // validation succeeded
          this.element.classList.remove('error');
          return resolve(result);
        }
        // validation failed
        this.element.classList.add('error');
        return reject(result);
      });
    });
  }
  /* DATE FUNCTIONS */

  dateInit() {
    if (this.value && !this.displayValue) this.dateDisplayValue();
  }
  dateDisplayValue() {
    if (typeof this.value === 'undefined' || this.value === '' || this.value === '0000-00-00') return this.value = null;
    this.displayValue = !this.value || this.value === '' || this.value === '0000-00-00' ? null : (/^\d{4}-\d{2}-\d{2}$/g).test(this.value) ? moment(this.value,'YYYY-MM-DD').format('DD/MM/YYYY') : this.value;
  }

  /* CHECKBOX FUNCTIONS */

  checkClick($event) {
    if (this.readonly) {
      $event.preventDefault();
      $event.stopPropagation();
    }
    else {
      this.value = !this.value;
      this.validate();
    }
  }

  /* EDITOR FUNCTIONS */

  editorInit() {
    if (this._editor_initialized) return;
    let config = {
      toolbar: ['heading', '|', 'bold', 'italic', 'link', 'bulletedList', 'numberedList', '|', 'indent', 'outdent', '|', 'blockQuote', 'insertTable', 'mediaEmbed', 'undo', 'redo'],
      heading: {
        options: [{
          model: 'paragraph',
          view: 'p',
          title: 'Paragrafo',
          class: 'ck-heading_paragraph'
        }, {
          model: 'heading1',
          view: 'h2',
          title: 'Intestazione 1',
          class: 'ck-heading_heading1'
        }, {
          model: 'heading2',
          view: 'h3',
          title: 'Intestazione 2',
          class: 'ck-heading_heading2'
        }, {
          model: 'heading3',
          view: 'h4',
          title: 'Intestazione 3',
          class: 'ck-heading_heading3'
        }]
      }
    };
    ClassicEditor.create(this.element.querySelector('.ka-input-editor'), config).then(editor => {
      editor.data.set(this.value || '');
      editor.isReadOnly = this.readonly;
      editor.model.document.on('change:data', () => { this.value = editor.data.get(); });
      this.editor = editor;
    }).catch(error => {
      konsole.error('There was a problem initializing ckeditor', error );
    }).finally(() => {
      this._editor_initialized = true;
    });
  }

  /* FILE FUNCTIONS */

  fileInit() {

  }
  fileDisplayValue() {
    if (typeof this.value === 'undefined' || this.value === '' || (this.value && !this.value.length)) return this.value = null;
    if (this.value === null) {
      // Hard reset field
      let form = document.createElement('form');
      form.style.display = 'none';
      this.element.parentElement.insertBefore(form, this.element);
      form.append(this.element);
      form.reset();
      form.parentElement.insertBefore(this.element, form);
      form.remove();
    }
    this.displayValue = this.value && this.value.length && this.value[0] && this.value[0].name ? this.value[0].name : this.value;
  }

  /* COMBO FUNCTIONS */

  comboInit() {
    if (this.value && !this.displayValue) this.comboDisplayValue();
  }
  comboOpen() {
    if (this.readonly) return;
    if (this.comboOpened) return this.comboClose();
    // Add backdrop
    if (!this.backdrop) {
      this.backdrop = document.createElement('div');
      this.backdrop.classList.add('ka-input-backdrop');
      this.backdrop.setAttribute('id', 'ka-input-backdrop-' + this.field);
      this.backdrop.addEventListener('click', event => { if (event.target == this.backdrop) this.comboClose(); });
    }
    this.backdrop.style.display = 'block';
    document.body.insertBefore(this.backdrop, document.body.children[0]);
    // Open combo
    this.comboOpened = true;
    this.comboItems = [];
    setTimeout(() => {
      // Move drawer to backdrop and open
      let bounds = this.drawer.getBoundingClientRect();
      let max_h = this.backdrop.getBoundingClientRect().height;
      this.backdrop.appendChild(this.drawer);
      if (bounds.top + 280 > max_h) {
        this.drawer.style.top = 'auto';
        this.drawer.style.bottom = ((max_h - bounds.top - 30) + 60) + 'px';
      }
      else {
        this.drawer.style.top = bounds.top + 'px';
        this.drawer.style.bottom = 'auto';
      }
      this.drawer.style.left = bounds.left + 'px';
      this.drawer.style.width = bounds.width + 'px';
      this.comboReady = true;
      this.comboQuery();
    }, 200);
  }
  comboClose() {
    // Move drawer back to component
    this.drawer.style = {};
    this.combo.appendChild(this.drawer);
    // Remove backdrop
    this.backdrop.style.display = 'none';
    // Close combo
    this.comboOpened = false;
    this.comboReady = false;
    this.comboQueryValue = null;
  }
  comboQuery(query) {
    this.comboGetItems(query).then((items) => {
      this.comboItems = items;
      this.comboParseSelected();
    });
  }
  comboGetItems(query, value) { // Returns combo items by matching text query or value
    return new Promise((resolve, reject) => {
      let items = [];
      let client = !this.client || this.datasource.url ? this._client : this.client;
      let datasource = null;

      if (typeof this.datasource === 'string' && (this.datasource.indexOf('{') === 0 || this.datasource.indexOf('[') === 0))
        datasource = JSON.parse(this.datasource);
      else if (typeof this.datasource === 'string' && this.parentContext.datasources && this.parentContext.datasources[this.datasource])
        datasource = this.parentContext.datasources[this.datasource];
      else if (Array.isArray(this.datasource))
        datasource = this.datasource;
      else datasource = this.datasource;
      
      if (this.datapreload && (!this.comboItemsSource || (!query && !value)) && (datasource.table || datasource.url)) {
        if (this.comboItemsSource) {
          items = this.comboItemsSource;
          if (query) this.comboGetItems(query).then((items) => { resolve(items); });
          if (value) this.comboGetItems(null, value).then((items) => { resolve(items); });
          else resolve(items);
        }
        else client.get(datasource.table || datasource.url, { distinct: true }).then(x => {
          items = this.comboItemsSource = this.parse_response(x);
          if (query) this.comboGetItems(query).then((items) => { resolve(items); });
          if (value) this.comboGetItems(null, value).then((items) => { resolve(items); });
          else resolve(items);
        }).catch(x => {});
      }
      else if (this.datapreload && (!this.comboItemsSource || (!query && !value)) && Array.isArray(datasource)) {
        items = this.comboItemsSource = datasource;
        if (!query && !value) resolve(items);
      }

      if (this.datapreload && this.comboItemsSource && query) {
        items = this.comboItemsSource.filter(x => { return x[this.datatext].toLowerCase().indexOf(query.toLowerCase()) >= 0; });
        resolve(items);
      }
      else if (this.datapreload && this.comboItemsSource && value) {
        items = this.comboItemsSource.filter(x => { return String(x[this.datavalue]) == String(value); });
        resolve(items);
      }
      else if (!this.datapreload && !query && !value && (datasource.table || datasource.url)) {
        resolve(items);
      }
      else if (!this.datapreload && query && (datasource.table || datasource.url)) {
        let endpoint = datasource.table || datasource.url;
        if (datasource.query && typeof datasource.query === 'object' && datasource.query.search) {
          endpoint = this.addUrlParam(endpoint, datasource.query.search, query);
        } else if (datasource.query && typeof datasource.query === 'string') {
          endpoint = this.addUrlParam(endpoint, datasource.query, query);
        } else {
          /* parsare e aggiungere i filters */
          let params = this.getUrlParams(endpoint);
          let datatext = this.datatext;
          let fields = this.datatext;
          let sort = `+${datatext}`;
          let keys = datatext.match(/\{[a-zA-Z_]*?\}/g);
          if (keys) {
            keys = keys.map(b=>b.replace(/\{|(.*?)\}/g, '$1'));
            datatext = keys[0];
            fields = keys.join(',');
            sort = '+' + keys.join(',+');
          }
          let operator = this.queryOperator ? this.queryOperator : '^=';
          operator = this.dataoperator ? this.dataoperator : operator;
          params.filters = (params.filters ? '&' : '') + `(${datatext}${operator}${query})`;
          params.sort = sort;
          params.fields = `${this.datavalue},${fields}`;
          endpoint = endpoint.replace(/\?.*$/, '') + '?' + this.buildQueryString(params);
          // endpoint += '?filters='+encodeURIComponent(this.datatext+'~'+query);
        }
        if (this.clientRequest && this.clientRequest.cancel) { this.clientRequest.cancel(); }
        this.clientRequest = client.get(endpoint, { distinct: true });
        this.clientRequest.then(x => {
          items = this.parse_response(x);
          resolve(items);
        }).catch(x => {});
      }
      else if (!this.datapreload && value && (datasource.table || datasource.url)) {
        let endpoint = datasource.table || datasource.url;
        if (datasource.query && typeof datasource.query === 'object' && datasource.query.item) {
          endpoint = this.addUrlParam(endpoint, datasource.query.item, value);
        } else if (datasource.query && typeof datasource.query === 'string') {
          endpoint = this.addUrlParam(endpoint, datasource.query, value);
        } else {
          /* parsare e aggiungere i filters */
          let params = this.getUrlParams(endpoint);
          let fields = this.datatext;
          let keys = fields.match(/\{[a-zA-Z_]*?\}/g);
          if (keys) fields = keys.join(',').replace(/\{|\}/g, '');
          params.filters = (params.filters ? '&' : '') + `(${this.datavalue}=${value})`;
          params.fields = `${this.datavalue},${fields}`;
          endpoint = endpoint.replace(/\?.*$/, '') + '?' + this.buildQueryString(params);
          // endpoint += '?filters='+encodeURIComponent(this.datavalue+'='+value);
        }
        //if (this.clientRequest && this.clientRequest.cancel) { this.clientRequest.cancel(); }
        this.clientRequest = client.get(endpoint, { distinct: true });
        this.clientRequest.then(x => {
          items = this.parse_response(x);
          resolve(items);
        }).catch(x => {});
      }
    });
  }
  comboDisplayValue() {
    if (!this.value) {
      if (this.buttons && this.buttons.length > 0 && this.control === 'combo' && this.datasource && this.datasource.table && !this.datamultiple) {
        this.buttons.openEnabled = false;
        this.buttons.addEnabled = !this.element.classList.contains('no-add');
      }
      return this.displayValue = null;
    }
    if (this.buttons && this.buttons.length > 0 && this.control === 'combo' && this.datasource && this.datasource.table && !this.datamultiple) {
      this.buttons.openEnabled = !this.element.classList.contains('no-open');
      this.buttons.addEnabled = false;
    }
    if (this.datamultiple) { // Display comma separated list of selected items datatext
      let datatexts = [];
      let parsed = 0;
      if (!this.value.length) {
        this.comboValueReady();
        return this.displayValue = null;
      }
      for (let item of this.value) {
        parsed ++;
        // Check if value is among combo items
        let v = typeof item === 'string' || typeof item === 'number' ? String(item) : String(item[this.datavalue]) ? String(item[this.datavalue]) :  null;
        this.comboGetItems(null, v).then((items) => {
          if (items.length === 1) { // Item found
            if (typeof item === 'object') item = v;
            let keys = this.datatext.match(/\{[a-zA-Z_]*?\}/g);
            if (keys) {
              let datatext = this.datatext;
              for (let key of keys) {
                datatext = datatext.replace(key, items[0][key.replace(/\{|\}/g, '')] || '');
              }
              datatexts.push(datatext);
            }
            else datatexts.push( items[0][this.datatext] ? items[0][this.datatext] : eval(`items[0].${this.datatext}`) ? eval(`items[0].${this.datatext}`) : 'undefined' );
          }
          else { // Item not found, remove from value
            item = null;
          }
          if (this.value.length === parsed) {
            this.displayValue = datatexts.join(', ');
          }
          this.comboValueReady();
        });
      }
    }
    else {
      if (this.comboItems) {
        // Check if value is among combo items
        let items = this.comboItems.filter(x => { return String(x[this.datavalue]) == String(this.value); });
        if (items && items[0]) {
          this.objectValue = items[0];
          let keys = this.datatext.match(/\{[a-zA-Z_]*?\}/g);
          if (keys) {
            let datatext = this.datatext;
            for (let key of keys) {
              datatext = datatext.replace(key, items[0][key.replace(/\{|\}/g, '')] || '');
            }
            this.displayValue = datatext;
          }
          else this.displayValue = items[0][this.datatext] ? items[0][this.datatext] : eval(`items[0].${this.datatext}`) ? eval(`items[0].${this.datatext}`) : 'undefined';
          this.comboValueReady();
          return;
        }
      }
      let v = typeof this.value === 'string' || typeof this.value === 'number' ? String(this.value) : String(this.value[this.datavalue]) ? String(this.value[this.datavalue]) : null;
      this.comboGetItems(null, v).then((items) => {
        if (items.length === 1) { // Item found
          this.objectValue = items[0];
          if (typeof this.value === 'object') this.value = v;
          let keys = this.datatext.match(/\{[a-zA-Z_]*?\}/g);
          if (keys) {
            let datatext = this.datatext;
            for (let key of keys) {
              datatext = datatext.replace(key, items[0][key.replace(/\{|\}/g, '')] || '');
            }
            this.displayValue = datatext;
          }
          else this.displayValue = items[0][this.datatext] ? items[0][this.datatext] : eval(`items[0].${this.datatext}`) ? eval(`items[0].${this.datatext}`) : 'undefined';
          this.comboValueReady();
        }
        else {
          this.value = null;
          this.displayValue = null;
          this.objectValue = null;
          this.comboValueReady();
        }
      });
    }
  }
  comboValueReady() {
    if (!this._ready) {
      this._ready = true;
      if (typeof this.ready === 'function') this.ready(this);
    }
  }
  comboSelect(item, $event) {
    // Combo a selezione singola
    if (!this.datamultiple) {
      this.value = String(item[this.datavalue]);
      this.comboParseSelected();
      this.comboClose();
      this.validate();
    }
    // Combo a selezione multipla
    else {
      if (!this.value || !Array.isArray(this.value)) this.value = [];
      // If item is already selected, deselect it
      if (this.value.length && this.value.indexOf(String(item[this.datavalue])) >= 0)
        this.value.splice(this.value.indexOf(String(item[this.datavalue])), 1);
      else
        this.value.push(String(item[this.datavalue]));

      this.comboParseSelected();
      this.validate();
    }
  }
  comboParseSelected() {
    // Deselect all
    for (let item of this.comboItems) item.selected = false;
    if (!this.value) return;
    // Set selection for datamultiple
    if (this.datamultiple) {
      for (let value of this.value) {
        let item = this.comboItems.filter(x => { return String(x[this.datavalue]) === String(value); });
        if (item && item[0]) item[0].selected = true;
      }
    }
    // Set selection for single value
    else {
      let item = this.comboItems.filter(x => { return String(x[this.datavalue]) === String(this.value); });
      if (item && item[0]) item[0].selected = true;
    }
    this.comboDisplayValue();
  }

  /**
   * COMBOLIST
   * @author Piero Vettori <p.vettori@mamoka.com>
   */
  @bindable() comboListItems = [];
  comboListAddItem(item, $event) {
    this.comboListItems.push({ value: item[this.datavalue], text: item[this.datatext] });
    this.comboListItems.sort((a, b) => { if (a.text < b.text) return -1; if (a.text > b.text) return 1; return 0; });
    this.value = this.comboListItems.map(x => x.value);
    console.info('comboListAddItem', this.comboListItems, this.value);
    this.validate();
    this.comboClose();
  }
  comboListRemoveItem(item, $event) {
    this.comboListItems.splice(this.comboListItems.indexOf(item), 1);
    this.value = this.comboListItems.map(x => x.value);
    console.info('comboListRemoveItem', this.comboListItems, this.value);
    this.validate();
  }
  comboListUpdate() {
    if (!this.value) return;
    if (!Array.isArray(this.value)) this.value = [this.value];
    this.value.forEach(item => {
      let listItem = { value: null, text: null };
      if (['number', 'string'].indexOf(typeof(item)) > -1) listItem = { value: item, text: item };
      else if (item[this.datavalue] && item[this.datatext]) listItem = { value: item[this.datavalue], text: item[this.datatext] };
      else if (item.value && item.text) listItem = Object.assign({}, item);
      if (this.comboListItems.filter(x => x.value === listItem.value).length === 0) this.comboListItems.push(listItem);
    });
    this.comboListItems.sort((a, b) => { if (a.text < b.text) return -1; if (a.text > b.text) return 1; return 0; });
  }


  /* CHECKLIST FUNCTIONS */

  checklistInit() {
    if (Array.isArray(this.datasource)) {
      this.checklistItems = JSON.parse(JSON.stringify(this.datasource));
      this.checklistParseSelected();
    }
    else {
      let client = !this.client || this.datasource.url ? this._client : this.client;
      client.get(this.datasource.table || this.datasource.url).then(x => {
        this.checklistItems = this.parse_response(x);
        this.checklistParseSelected();
      }).catch(x => {});
    }
  }
  checklistSelect(item, $event) {
    if (this.readonly) return;
    if (!this.value || !Array.isArray(this.value)) this.value = [];
    // If item is already selected, deselect it
    if (this.value.length && this.value.indexOf(String(item[this.datavalue])) >= 0)
      this.value.splice(this.value.indexOf(String(item[this.datavalue])), 1);
    else
      this.value.push(String(item[this.datavalue]));

    this.checklistParseSelected();
    this.validate();
  }
  checklistParseSelected() {
    if (!this.checklistItems) return;// this.checklistInit();
    // Deselect all
    for (let item of this.checklistItems) item.selected = false;
    if (!this.value) return;
    for (let value of this.value) {
      let item = this.checklistItems.filter(x => { return String(x[this.datavalue]) == String(value); });
      if (item && item[0]) item[0].selected = true;
    }
  }

  parse_response(msg) {
    if (msg.responseType === 'json' && typeof msg.response === 'string') msg.response = JSON.parse(msg.response);
    return msg.response;
  }

  getUrlParams(url) {
    url = url.replace(/^.*(\?|$)/, '');
    if (!url[1]) return {};
    let chunks = url[1].split('&');
    return chunks.reduce((params, hash) => {
      let [key, val] = hash.split('=');
      return Object.assign(params, { [key]: decodeURIComponent(val) });
    }, {});
  }
  addUrlParams(url, params) {
    if (!params) return url;
    for (let [k, v] of Object.entries(params)) {
      url = this.addUrlParam(url, k, v);
    }
    return url;
  }
  addUrlParam(url, key, value) {
    if (!key || !value) return url;
    url = new URL((!url.startsWith('http') ? 'void://' : '') + url);
    let urlParams = new URLSearchParams(url.search);
    urlParams.append(key, value);
    url.search = urlParams;
    return url.toString().replace('void://', '');
  }

  buildQueryString(params) {
    return Object.keys(params).reduce((query, key) => {
      return `${query}&${key}=` + encodeURIComponent(params[key]);
    }, '').replace(/^\&/, '');
  }
}


export class kaHighlightValueConverter {
  toView(item, datatext, string) {
    let value = null;
    let keys = datatext.match(/\{[a-zA-Z_]*?\}/g);
    if (keys) {
      for (let key of keys) {
        datatext = datatext.replace(key, item[key.replace(/\{|\}/g, '')] || '');
      }
      value = datatext;
    }
    else value = item[datatext] ? item[datatext] : eval(`item.${datatext}`) ? eval(`item.${datatext}`) : null;
    if (!value) return;
    if (!string || !string.length) return value;
    let index = value.toLowerCase().indexOf(string.toLowerCase());
    if (index >= 0) value = value.substring(0, index) + '<strong>' + value.substring(index, index + string.length) + '</strong>' + value.substring(index + string.length);
    return value;
  }
}
