import {bindable, bindingMode, customElement, inject} from 'aurelia-framework';
import {Router} from 'aurelia-router';
import {KaToast} from '../ka-toast/ka-toast';
import {KaDialog} from '../ka-dialog/ka-dialog';

@customElement('ka-resource')
@inject(Element, Router, KaToast, KaDialog)
export class KaResource {
  @bindable() resource = null;
  @bindable() data = {};
  @bindable() schema = {};
  @bindable() injectData = {};
  @bindable() injectSchema = {};
  @bindable() client = null;
  @bindable() endpoint = null;
  @bindable() disabled = false;
  @bindable() controls = {};
  @bindable() listRefQuery = null;
  @bindable({ defaultBindingMode: bindingMode.fromView }) mode = 'normal'; // 'normal', 'edit', 'readonly'

  readonly = true;
  loaded = false;
  saveAndReload = false;
  errors = [];
  hasBeenEdited = false;

  // Callbacks - Keep them for backwards compatibility
  schema_loaded = () => {};
  data_loaded = () => {};
  save_before = (data) => { return data; };
  save_after = (data) => { return data; };

  buttonsActions = {
    prev: () => { return this.prev(); },
    next: () => { return this.next(); },
    close: () => { return this.close(); },
    edit: () => { return this.edit(); },
    save: () => { return this.save(); },
    delete: () => { return this.delete(); }
  };

  constructor(element, router, toast, dialog) {
    this.element = element;
    this.router = router;
    this.toast = toast;
    this.dialog = dialog;

    // CUSTOM EVENTS
    this.events = document.createTextNode(null);
    this.events.schemaLoaded = new CustomEvent('schemaLoaded', { resource: this });
    this.events.dataLoaded = new CustomEvent('dataLoaded', { resource: this });
    this.events.dataValidated = new CustomEvent('dataValidated', { resource: this });
    this.events.dataValidateError = new CustomEvent('dataValidateError', { resource: this });
    this.events.dataSaved = new CustomEvent('dataSaved', { resource: this });
    this.events.dataSaveError = new CustomEvent('dataSaveError', { resource: this });
  }

  bind(bindingContext, overrideContext) {
    this.parentContext = bindingContext;
    this.parentContext.resource = this;
  }

  attached() {
    this.load();
  }

  modeChanged(value) {
    switch (value) {
    case 'edit':
      this.disabled = false;
      this._setReadonly(false);
      break;
    case 'readonly':
      this.disabled = true;
      this._setReadonly(true);
      break;
    default:
      this.disabled = false;
      this._setReadonly(true);
      break;
    }
  }

  reload(resource) {
    this.data = {};
    //this.schema = {};
    this.readonly = true;
    this.loaded = false;
    this.errors = [];
    this.resource = resource;
    this.load();
  }
  load() {
    if (this.parentContext.params && this.parentContext.params.id && this.router.currentInstruction.params && this.router.currentInstruction.params.id && this.parentContext.params.id !== this.router.currentInstruction.params.id) { this.saveAndReload = true; }
    if (!this.endpoint && this.parentContext.endpoint) this.endpoint = this.parentContext.endpoint;
    if (!this.client && this.parentContext.api && this.parentContext.api.client) this.client = this.parentContext.api.client;
    if (!this.resource && this.parentContext.params && this.parentContext.params.id) this.resource = { id: this.parentContext.params.id };
    if (!this.listRefQuery && this.parentContext.params && this.parentContext.params.listRefQuery) this.listRefQuery = atob(this.parentContext.params.listRefQuery);
    this.modeChanged(this.mode);

    if (this.client && this.endpoint) {
      let endpoint = this.endpoint;
      let dataset = this.endpoint;//.replace(/\/.*$/, '');
      this._request('get', `datasets/${dataset}`).then(response => {
        let schema = {};
        for (let item of response) schema[item.field] = item;
        this.schema = schema;
        this.schema_loaded(); // old way
        this.events.dispatchEvent(this.events.schemaLoaded); // new way

        if (this.resource && this.resource.id === 'new') {
          this.originalData = {};
          for (let key of Object.keys(this.schema)) this.originalData[key] = null;
          this.data = JSON.parse(JSON.stringify(this.originalData));
          if (this.parentContext.params && this.parentContext.params.data) this.data = Object.assign(this.data, this.parentContext.params.data);
          this.loaded = true;
          this.data_loaded(); // old way
          this.events.dispatchEvent(this.events.dataLoaded); // new way
          this._setReadonly(false);
        } else if (this.resource && this.resource.id) {
          this._request('get', `${endpoint}/${this.resource.id}?depth=0`).then(data => {
            this.originalData = data;
            this.data = JSON.parse(JSON.stringify(this.originalData));
            // Should be wrong to overwrite loaded record data with defalut data passed through params
            //if (this.parentContext.params && this.parentContext.params.data) this.data = Object.assign(this.data, this.parentContext.params.data);
            this.loaded = true;
            this.data_loaded(); // old way
            this.events.dispatchEvent(this.events.dataLoaded); // new way
          });
        }
      });
    }
    else if (this.injectSchema && this.injectData) {
      // Parse passed schema
      let schema = {};
      for (let item of this.injectSchema) schema[item.field] = item;
      this.schema = schema;
      // Parse passed data
      this.originalData = this.injectData;
      this.data = JSON.parse(JSON.stringify(this.originalData));
      if (this.parentContext.params && this.parentContext.params.data) this.data = Object.assign(this.data, this.parentContext.params.data);
      setTimeout(()=>{
        this.loaded = true;
        this.data_loaded(); // old way
        this.events.dispatchEvent(this.events.dataLoaded); // new way
      }, 500);
    }
  }

  prev() {
    let listRefQuery = this.listRefQuery.replace(/offset=(\d+)/, (match, p1) => { return `offset=${parseInt(p1, 10) === 0 ? 0 : parseInt(p1, 10) - 1}`; });
    this.client.get(`${listRefQuery}`).then(x => {
      let response = x.response;
      if (!response.length) return this.toast.show('Non ci sono record precedenti a questo', 'error');
      if (this.parentContext.params.dialog) {
        this.listRefQuery = listRefQuery;
        this.reload(response[0]);
      } else {
        let route = this.router.currentInstruction.config;
        this.router.navigateToRoute(route.name, { id: response[0].id, listRefQuery: btoa(listRefQuery) }, { trigger: true, replace: true });
      }
    }).catch(() => { this.toast.show('Errore nell\'apertura del record precedente', 'error'); });
  }

  next() {
    let listRefQuery = this.listRefQuery.replace(/offset=(\d+)/, (match, p1) => { return `offset=${parseInt(p1, 10) + 1}`; });
    this.client.get(`${listRefQuery}`).then(x => {
      let response = x.response;
      if (!response.length) return this.toast.show('Non ci sono record successivi a questo', 'error');
      if (this.parentContext.params.dialog) {
        this.listRefQuery = listRefQuery;
        this.reload(response[0]);
      } else {
        let route = this.router.currentInstruction.config;
        this.router.navigateToRoute(route.name, { id: response[0].id, listRefQuery: btoa(listRefQuery) }, { trigger: true, replace: true });
      }
    }).catch(() => { this.toast.show('Errore nell\'apertura del record successivo', 'error'); });
  }

  delete() {
    this.dialog.open({
      type: 'confirm',
      class: 'warning',
      title: 'Elimina',
      body: 'Eliminare definitivamente l\'elemento?'
    }, true).whenClosed(response => {
      if (!response.wasCancelled) {
        this._request('delete', `${this.endpoint}/${this.resource.id}`).then(() => {
          if (this.parentContext.params.dialog)
            this.parentContext.params.dialog.close(true);
          else
            this.router.navigate(this.router.currentInstruction.config.href);
        });
      }
    });
  }

  edit() {
    if (this.endpoint) {
      this._setReadonly(false);
    }
  }

  save() {
    this.errors = this._validateControls().then(success => {
      this.events.dispatchEvent(this.events.dataValidated);
      if (this.endpoint) {
        let promises = [];
        let data = this.save_before(this.data);
        if (data === false) return this.data = data;
        this.data = data;
        data = {};
        for (let [field, item] of Object.entries(this.data)) {
          promises.push(new Promise((resolve, reject) => {

            // Skip unchanged fields when updating record
            if (this.resource.id && this.resource.id !== 'new' && this.originalData[field] === item) resolve(true);

            // FILE
            else if (this.controls[field] && (this.controls[field].control === 'file' || (this.controls[field].schema && this.controls[field].schema.control === 'file') ) && item && item.length && item[0] && item[0].name) {
              let filedata = {};
              filedata.filename = item[0].name;
              let reader = new FileReader();
              reader.readAsDataURL(item[0]);
              reader.onload = (x) => {
                filedata.stream = reader.result.indexOf('base64,') ? reader.result.slice(reader.result.indexOf('base64,') + 7) : reader.result;
                data[field] = filedata;
                resolve(true);
              };
            }
            else if (this.controls[field] && (this.controls[field].datatype === 'json' || (this.controls[field].schema && this.controls[field].schema.control === 'json') ) && item && typeof item === 'string') {
              data[field] = JSON.parse(item);
              resolve(true);
            }
            // PASSWORD
            /*
            else if (this.controls[field] && this.controls[field].control == 'password' && item === null) {
              // We do not send password at all
              resolve(true);
            }
            */
            // EDITOR
            else if (this.controls[field] && (this.controls[field].control === 'editor' || (this.controls[field].schema && this.controls[field].schema.control === 'editor') ) && item === '') {
              data[field] = null;
              resolve(true);
            }
            else {
              if (typeof item === 'undefined') console.error(`Input ${field} has undefined value. Skipping...`);
              else data[field] = JSON.parse(JSON.stringify(item));
              resolve(true);
            }
          }));
        }
        Promise.all(promises).then(() => {
          let method = 'post';
          let endpoint = this.endpoint;//.replace(/\/.*$/, '');
          if (this.resource.id && this.resource.id !== 'new') {
            delete(data.id);
            method = 'patch';
            endpoint = `${endpoint}/${this.resource.id}`;
          }
          // Remove fields on which user has no permission for post/patch
          let user = this.parentContext.api && this.parentContext.api.user ? this.parentContext.api.user : null;
          if (user && user.groups && user.groups.length) {
            for (let [key, value] of Object.entries(data)) {
              if (!this.schema[key] || !this.schema[key][method]) delete(data[key]);
              let permissions = this.schema[key][method];
              let granted = Object.keys(permissions).filter((k) => {
                return (k === '0' || user.groups.includes(k)) && permissions[k] === '1';
              });
              if (!granted || !granted.length) delete(data[key]);
            }
          }
          if (method === 'patch') {
            data = JSON.parse(JSON.stringify(data));
            if (!data || !Object.keys(data).length) {
              // No data to patch, back to readonly
              this._setReadonly(true);
              this.events.dispatchEvent(this.events.dataSaved); // new way
              return;
            }
          }
          this.hasBeenEdited = true;
          this._request(method, endpoint, data).then(response => {
            // console.info('response', response);
            this._setReadonly(true);
            data = Object.assign(this.originalData, data);
            data.id = response.id || this.resource.id;
            this.data = this.save_after(data);
            this.events.dispatchEvent(this.events.dataSaved); // new way
            this.originalData = JSON.parse(JSON.stringify(data));
            if (this.data === false) return this.data = data;

            let route = this.router.currentInstruction.config;
            if (this.resource.id === 'new') {
              this.toast.show('Record creato!', 'success');
              if (this.parentContext.params.dialog || this.saveAndReload) {
                //this.parentContext.params.dialog.close(true, response);
                this.reload(response);
              } else {
                this.router.navigateToRoute(route.name, { id: response.id || this.resource.id, listRefQuery: (this.listRefQuery ? btoa(this.listRefQuery) : null) }, { trigger: true, replace: true }); //this.router.navigate(route.href);
              }
            } else {
              this.toast.show('Record aggiornato!', 'success');
              if (this.parentContext.params.dialog || this.saveAndReload) {
                //this.parentContext.params.dialog.close(true, response);
                this.resource.id = response.id || this.resource.id;
                this.modeChanged(this.mode);
              } else {
                this.router.navigateToRoute(route.name, { id: response.id || this.resource.id, listRefQuery: (this.listRefQuery ? btoa(this.listRefQuery) : null) }, { trigger: true, replace: true });
              }
            }
          }, error => {
            console.warn('error', error);
            error = JSON.parse(error.response);
            this.events.dispatchEvent(this.events.dataSaveError); // new way
            if (error.errors) {
              for (let err of error.errors) {
                // Gestisci errore formato kamaji
                if (err.arg && Array.isArray(err.arg) && err.arg[0]) this.toast.show(`<strong>${this.schema[err.arg[0]].label}</strong>: ${err.msg}`, 'error');
                else if (err.arg && typeof err.arg === 'object') this.toast.show(`<strong>${this.schema[Object.keys(err.arg)[0]].label}</strong>: ${err.msg}`, 'error');
                // Gestisci errore formato gateway
                else if (err.message) {
                  let message = (err.data && this.schema[err.data] ? `<strong>${this.schema[err.data].label}</strong>: ` : '') + err.message;
                  this.toast.show(message, 'error');
                }
              }
            }
          });
        });
      }
    }, response => {
      // console.warn('validation errors', response);
      this.events.dispatchEvent(this.events.dataValidateError);
      if (response.length) {
        for (let errors of response) {
          for (let error of errors) {
            if (!error.valid) this.toast.show(`<strong>${error.object.label}</strong>: ${error.message}`, 'error');
          }
        }
      }
    });
  }

  close() {
    let route = this.router.currentInstruction.config;
    if (this.readonly || this.resource.id === 'new') {
      if (this.parentContext.params.dialog) {
        let ok = this.hasBeenEdited === true;
        this.parentContext.params.dialog.close(ok, ok && this.originalData ? this.originalData : null);
        return;
      }
      this.router.navigate(route.href);
      return;
    }
    this.data = JSON.parse(JSON.stringify(this.originalData));
    this._setReadonly(true);
  }

  _validateControls() {
    for (let index in this.tabs) this.tabs[index].errors = [];
    let promises = [];
    for (let [field, item] of Object.entries(this.controls)) {
      let promise = item.validate().then(result => {
        if (!result.valid) {
          let errors = [];
          for (let error of item.validator.errors) {
            error.tab = item.parentTab;
            //this[item.parentTab].errors.push(error);
            errors.push(error);
          }
          return Promise.reject(errors);
        }
        return Promise.resolve(result);
      });
      if (item.validator) promises.push(promise);
    }
    return this.promiseAll(promises);
  }

  _resetButtons() {
    let buttons = [];
    if ((this.disabled || this.readonly) && this.listRefQuery) {
      buttons.push({ action: () => this.buttonsActions.prev(), label: 'Record precedente', icon: 'keyboard_arrow_left' });
      buttons.push({ action: () => this.buttonsActions.next(), label: 'Record seguente', icon: 'keyboard_arrow_right' });
    }
    if (this.client && !this.disabled) {
      if ((!this.resource.id || this.resource.id !== 'new') && this.readonly) {
        buttons.push({ action: () => this.buttonsActions.delete(), label: 'Elimina',  icon: '&#xE92B;' });
      }
      if (this.readonly) {
        buttons.push({ action: () => this.buttonsActions.edit(),   label: 'Modifica', icon: '&#xE3C9;' });
      } else {
        buttons.push({ action: () => this.buttonsActions.save(),   label: 'Salva',    icon: '&#xE5CA;' });
      }
    }
    buttons.push({ action: () => this.buttonsActions.close(), label: 'Chiudi', icon: '&#xE14C;' });
    this.buttons = buttons;
  }

  _setReadonly(value) {
    this.readonly = value;
    this._resetButtons();
  }

  _request(method, url, data = {}) {
    if (!this.client) return Promise.reject('Missing HTTP client.');
    method = method.toLowerCase();
    if (!this.client[method]) return Promise.reject('Unknown HTTP method.');
    return this.client[method](url, data).then(x => {
      return Promise.resolve(x.response);
    }).catch(x => {
      return Promise.reject(x);
    });
  }


  promiseAll(promises) {
    let count = 0;
    let errors = [];
    let results = [];
    let promise = new Promise((resolve, reject) => {
      for (let p of promises) {
        p.then(success => {
          count++;
          results.push(success);
          if (count === promises.length) {
            if (errors.length) reject(errors);
            else resolve(results);
          }
          return success;
        }, error => {
          count++;
          errors.push(error);
          if (count === promises.length) {
            if (errors.length) reject(errors);
            else resolve(results);
          }
          return error;
        });
      }
    });
    return promise;
  }

}
